//
// DOSPKT - MS-DOS based version of PacketScript coverted from the original
//          MAC source code.  
// 
// Copyright 1996 by Brian Barnt
//
// You may freely distribute this source code, and use all or part of
// this source code in all software including commercial, freeware,
// shareware and private applications.
//
// Please report bugs, fixes, etc to me at:
//   73457,1510 (compuserve)
//
// Created	05/30/96
//

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "bc_sio.h"
#include "PktScrpt.h"
#include "sender.h"
#include "dcc.h"

long flength(FILE *fp);
unsigned uPort,uDivisor,running;
unsigned char szConfigBuf[512], cOutputType;

//
// PrintMessage is the implementation-specific routine to display
// 'Print' statements.
//

void PrintMessage(char *szMessage)
{
   fprintf(stdout,"%s\n",szMessage);
}


//
// ReportParseError is the implementation specific routine to display
// error messages.
//

void ReportParseError(int errCode, long errLine, char *pErrLineText, long errPos)
{
	int		i;
	char	message[256];
	char	text[256];
	char	carret[256];
	
	sprintf(message, "Error %d: line %ld", errCode, errLine);
	
	strcpy(text, pErrLineText);
	
	carret[0] = '\0';
	for (i=0; i<errPos; i++)
		strcat(carret, " ");
	strcat(carret, "^");
	switch (errCode)
	{
		case FAILED_SYNTAX:		strcat(carret, "- Syntax error.");			break;
		case FAILED_VALUE:		strcat(carret, "- Bad value.");				break;
		case FAILED_MAKEPACKET:	strcat(carret, "- MakePacket failed.");		break;
		default:															break;
	}
	
	puts(message);
   puts(text);
   puts(carret);
}

//
// Ascii to Hex
//

int atoh(char *s)
{
   int i,l,r,v;
   char *b, buf[40];

   b = &buf[0];

   for(l=0;isxdigit(*s);l++)
      buf[l] = *s++;
   buf[l] = '\0';

   if (l < 4)
      {
      strrev(b);
      for (i=0;i<l;i++)
         {
         v = *b++ - '0';
         if (v > 9) v -= 7;
         if (i == 0) r = v;
         if (i == 1) r += (v * 16);
         if (i == 2) r += (v * 256);
         if (i == 3) r += (v * 4096);
         }
      }
   else
      r = 0;

   return(r);
}



//
// Calculate Divisor 
//

void CalcDivisor(char *s)
{
   while (*++s != ' ')     // Skip to end of Token
      ;
   while (*++s == ' ')     // Skip white space
      ;
   
   uDivisor = atoi(s);
   if (uDivisor == 0 || uDivisor >= 10)
      uDivisor = 6;
}

//
// Calculate Output Type
//

void CalcOutput(char *s)
{
   while (*++s != ' ')     // Skip to end of Token
      ;
   while (*++s == ' ')     // Skip white space
      ;

   if (strncmp(s,"SENDER",6) == 0)
      cOutputType = OT_SENDER;
   else
      cOutputType = OT_SERIAL;
}

//
// CalcPort calculates the base I/O address of the desired serial port.
//
//

void CalcPortAddress(char *s)
{
   while (*++s != ' ')
      ;
   while (*++s == ' ')
      ;

   uPort = 0;
   if (strncmp(s,"COM",3) == 0)
      {
      s += 3;
      switch (*s)
         {
         case '1':
            uPort = 0x03f8;    /* COM1  at 0x03f8            */
            break;
         case '2':
            uPort = 0x02f8;    /* COM2  at 0x02f8            */
            break;
         case '3':
            uPort = 0x3220;    /* COM3  at 0x3220            */
            break;
         case '4':
            uPort = 0x3228;    /* COM4  at 0x3228            */
            break;
         case '5':
            uPort = 0x4220;    /* COM5  at 0x4220            */
            break;
         case '6':
            uPort = 0x4228;    /* COM6  at 0x4228            */
            break;
         case '7':
            uPort = 0x5220;    /* COM7  at 0x5220            */
            break;                              
         case '8':
            uPort = 0x5228;    /* COM8  at 0x5228            */
            break;
         }
      }
   
   if (strncmp(s,"0X",2) == 0)  // address is in Hex
      {
      s += 2;
      uPort = atoh(s);
      }
   
   if (uPort == 0)
      uPort = 0x03f8;          /* default is COM 1           */
}


//
// ReadConfig reads the configuration file for the serial port.
//

void ReadConfigFile()
{
   FILE *fp;

   if ((fp = fopen("config.dcc","r")) == NULL)
      {
      printf("\nERROR: Could not open configuration file\n");
      exit (1);
      }

   while (!feof(fp))
      {
      if (fgets(szConfigBuf,255,fp) == NULL)
         return;
      strupr(szConfigBuf);
      szConfigBuf[strlen(szConfigBuf) - 1] = '\0';
      switch (szConfigBuf[0])
         {
         case 'P':
            if (szConfigBuf[1] == 'O')
               {
               CalcPortAddress(szConfigBuf);
               break;
               }
            break;
         case 'D':
            if (szConfigBuf[1] == 'I')
               {
               CalcDivisor(szConfigBuf);
               break;
               }
            break;
         case 'O':
            CalcOutput(szConfigBuf);
            break;
         }
      }
   fclose(fp);
}


void print_pkt(unsigned char *packetToPrint,long bytesInPacket)
{
   long  i;
   int   iChr;

   for (i = 0; i < bytesInPacket; i++)
      {
      printf(" 0x%2.2X",*packetToPrint++);
      if ((i > 0) && (( i % 25) == 0))
         {
         iChr = kb_getc();
         if (iChr == 0x1B)
               break;
         }

      }
}

//
// Main 
//

int main (int iArgCnt, char *pszArg[])
{
   FILE           *fpScript;
   unsigned char  *pText;
   long           lTextLen, lActualLen;
	short			   hit;
	long			   totalByteCount, totalTime;
	CodeElement    *pProgram;
	int			   done = FALSE;
	long			   t;

   fprintf(stderr,"DOSPKT - PacketScript for MS-DOS\nVersion 0.95\nOriginal Mac Version by Ken Rice\nSender Board Drivers by Ken West\nPorted to MS-DOS by Brian Barnt\nFebruary 16, 2001\n");

   if (iArgCnt < 2)
      {
      fprintf(stderr,"Useage: DOSPKT <script file>\n\n");
      fprintf(stderr,"Configuratuion File (CONFIG.DCC) Entries:\n");
      fprintf(stderr,"   OUTPUT  xxxxxx - SERIAL port or SENDER board output\n");
      fprintf(stderr,"   PORT    xxx    - Hexadecimal Address of Serial Pprt\n");
      fprintf(stderr,"   DIVISOR nn     - Baud Rate Divisor\n\n");
      return -1;
      }

    ReadConfigFile();

    if (cOutputType == OT_SERIAL)
      {
      printf("Initializing Serial Port\n");
      if (mdm_open(uPort,uDivisor,3) == 0xff)   //  8-N-1
         {
         printf("Error Initializing com port and interupt vector.\n");
         return -1;
         }
      }

   if (cOutputType == OT_SENDER)
      {
      init_send();
      if (start_clk() == FAIL)
         {
         printf("Error Initializing Sender Board.\n");
         return -1;
         }
      }

   if ((fpScript = fopen(pszArg[1],"rb")) == NULL)
      {
      fprintf(stderr,"Error opening %s for reading.\n",pszArg[1]);
      return -2;
      }

   if ((lTextLen = flength(fpScript)) == -1)
      {
      fprintf(stderr,"Error finding the length of %s.\n",pszArg[1]);
      fclose(fpScript);
      return -3;
      }

	pText = (unsigned char *) malloc(lTextLen);
   if (pText == NULL)
      {
      fprintf(stderr,"Could not allocate memory to read %s.\n",pszArg[1]);
      fclose(fpScript);
      return(-4);
      }

   if ((lActualLen = fread(pText,sizeof(char),lTextLen,fpScript)) != lTextLen)
      {
      fprintf(stderr,"Error reading %l bytes (%l actual) from %s.\n",
               lTextLen,lActualLen,pszArg[1]);
      free(pText);
      fclose(fpScript);
      return(-5);
      }

   fclose(fpScript);

   printf("\nParsing Script...");

 	if (ParseScript(pText, lTextLen,cOutputType) == SUCCESS)
   	{
      printf("Sucessful\n");

      pProgram = GetCode();

		/* Calculate and display total run time of script in microsecond. */

		totalByteCount = CountTotalBytesToSend(pProgram);
		totalTime = totalByteCount;
		totalTime *= 560;

		t = totalTime/1000000;
		printf("Total Bytes: %lu\nApprox. Run Time: %ld.%06ld\n",totalByteCount, t, totalTime - (t*1000000) );

      RunCode(pProgram,cOutputType);

      }


   free(pText);

   if (cOutputType == OT_SERIAL)
      {
      printf("\nReseting Serial Port");
      mdm_close();
      }

   if (cOutputType == OT_SENDER)
      {
      stop_clk();
      }

   return 0;
}

//
// flength returns the length of a file.  If an error occurs, -1 is returned.
//

long flength(FILE *fp)
{
   long lLength, lOriginal;

   // Save current file position

   lOriginal = ftell(fp);

   if (fseek(fp,0L,SEEK_END))  // Goto end of file
      return -1;

   lLength = ftell(fp);    // Save the length of the file

   if (fseek(fp,lOriginal,SEEK_SET)) // Restore the original position
      return -1;

   return lLength;

}
