#include "stdio.h"
#include "dos.h"
#include "sender.h"

extern unsigned uPort;
extern unsigned running;
extern int	clk0t;
extern int   clk0h;
extern int   clk1t;

extern const u_short	 	PORT_DEF;	 		// Default I/O base port.
extern const u_int		SAN_CNT;			// Sanity timeout value.
extern const u_int		ERR_MAX;				// Max contig. errors to print.
extern u_long				p_cnt;					// Count of packets sent.
extern u_long				b_cnt;					// Count of Bytes sent.
extern enum Scope_states	scope;					// Scope trigger state.
extern u_int				   err_cnt;				// Count of contiguous errors.
extern u_int				   m_pc_delay_1usec;	   	// Loop count for 1 usec. delay.
extern unsigned         do_crit;


/*
 *	Port B input bit positions.
 */
extern const BYTE	 	PB_POS_UNDER;			// Underflow.
extern const BYTE	 	PB_POS_CLK1L;			// Clk 1 NOT.
extern const BYTE	 	PB_POS_DCC0;			// DCC out 0.
extern const BYTE		PB_POS_DCCQH;			// DCC data out.
extern const BYTE		PB_POS_82G0H;			// 8254 0H gate.
extern const BYTE		PB_POS_SHOLDL;			// Shift hold NOT.
extern const BYTE		PB_POS_SCOPE;			// Scope output.
extern const BYTE		PB_POS_DCCOUTD;			// DCC signal output.

/*
 *	Port B input bit masks.
 */
extern const BYTE		PB_MSK_UNDER;		// Underflow.
extern const BYTE		PB_MSK_CLK1L;		// Clk 1 NOT.
extern const BYTE		PB_MSK_DCC0;		// DCC out 0.
extern const BYTE		PB_MSK_DCCQH;		// DCC data out.
extern const BYTE		PB_MSK_82G0H;		// 8254 0H gate.
extern const BYTE		PB_MSK_SHOLDL;		// Shift hold NOT.
extern const BYTE		PB_MSK_SCOPE;		// Scope output.
extern const BYTE		PB_MSK_DCCOUTD;		// DCC signal output.

/*
 *	Port C output bit values.
 */
extern const BYTE		PC_POS_BDRST; 			// Reset Board.
extern const BYTE		PC_POS_CPUEN;			// Enable CPU clock.
extern const BYTE		PC_POS_CPUCLKL;			// CPU clock NOT.
extern const BYTE		PC_POS_INTRA;			// Interrupt active.
extern const BYTE		PC_POS_UNDERCLRL;			// Clear under NOT.
extern const BYTE		PC_POS_SCOPENL;			// Enable Scope output NOT.
extern const BYTE		PC_POS_ACKAL;			// Data Ack. (Input)
extern const BYTE		PC_POS_OBFAL;			// Output buffer full NOT.

/*
 *	Port C output bit masks.
 */
extern const BYTE	  	PC_MSK_BDRST;		// Reset Board.
extern const BYTE	  	PC_MSK_CPUEN;		// Enable CPU clock.
extern const BYTE	  	PC_MSK_CPUCLKL;		// CPU clock NOT.
extern const BYTE	  	PC_MSK_INTRA;		// Interrupt active.
extern const BYTE	  	PC_MSK_UNDERCLRL;		// Clear under NOT.
extern const BYTE	  	PC_MSK_SCOPENL;		// Enable Scope output NOT.
extern const BYTE	  	PC_MSK_ACKAL;		// Data Ack. (Input)
extern const BYTE	  	PC_MSK_OBFAL;		// Output buffer full NOT.

/*
 *	8255 Control port values.
 */
extern const BYTE		P_CNTRL_RST;			// Power up (all inputs).
extern const BYTE		P_CNTRL_NORM;			// Normal mode.
													// A		-	Mode 1 out
													// B		-	Mode 0 in
													// C upper	-	Out
													// C lower	-	Out

/*
 *	8254 Control port values.
 */
extern const BYTE	   	CNTRL_0T_NORM;			// 0T normal (mode 3).
extern const BYTE	   	CNTRL_0H_RST;			// 0H reset (mode 0).
extern const BYTE	   	CNTRL_0H_NORM;			// 0H normal (mode 5).
extern const BYTE	   	CNTRL_1T_NORM;			// 1T normal (mode 3).

/*
 *	8254 counter values.
 */
extern const u_short  	CNT_0T_INIT;		// 0T initial count.
extern const u_short  	CNT_0H_INIT;	// 0H initial count.
extern const u_short  	CNT_1T_INIT;		// 1T initial count.


extern const u_int		DECODER_AVG_MIN;
extern const u_int		DECODER_AVG_NOM;
extern const u_int		DECODER_AVG_MAX;

extern const BYTE		CHECK_INIT;

/*
 *	Byte and packet rates per second.
 */
extern const u_int		ZERO_SEC_MIN;	// 0 bytes per sec @ MIN.
extern const u_int		ZERO_SEC_NOM;	// 0 bytes per sec @ NOM.
extern const u_int		ZERO_SEC_MAX;	// 0 bytes per sec @ MAX.
extern const u_int		ONE_SEC_MIN;		// 1 bytes per sec @ MIN.
extern const u_int		ONE_SEC_NOM;		// 1 bytes per sec @ NOM.
extern const u_int		ONE_SEC_MAX;		// 1 bytes per sec @ MAX.
extern const u_int		RESET_SEC_MIN;	// Reset pkts per sec @ MIN.
extern const u_int		RESET_SEC_NOM;	// Reset pkts per sec @ NOM.
extern const u_int		RESET_SEC_MAX;	// Reset pkts per sec @ MAX.
extern const u_int		IDLE_SEC_MIN;	// Idle pkts per sec @ MIN.
extern const u_int		IDLE_SEC_NOM;	// Idle pkts per sec @ NOM.
extern const u_int		IDLE_SEC_MAX;	// Idle pkts per sec @ MAX.
/*
 *	Speed related values.
 */
extern const int		SP_E_STOP;		// Emergency stop.
extern const int		SP_MIN;			// Min speed (stop).
extern const int		SP_MAX;			// Max speed.
extern const int		ACC_MAX;			// Max accessory id.

/*
 *	Address related values.
 */
extern const int		ADDR_MIN;		   // Min command address.
extern const int		ADDR_MAX;		   // Max command address.
extern const int		ACC_ADDR_MIN;	// Min accessory address.
extern const int		ACC_ADDR_MAX;	// Max accessory address.

/*
 *	Invalid bit time value.
 */
extern const u_int		T_INV;				   // Invalid time value.


void start_crit( void )
{
	if ( do_crit )
	   {
		_disable();
		OUT_PC( PC_POS_UNDERCLRL, 0 );
	   }
}
void	clr_under( void ) { OUT_PC( PC_POS_UNDERCLRL, 0 ); }

void	clr_err_cnt( void ) { err_cnt = 0; }


/*--------------------------------------------------------------------------*/
/*
 *	NAME
 *
 *		init_8254()							-	 Init 8254 timer.
 *
 *	RETURN VALUE
 *
 *		OK		-	Sucess.
 *		FAIL	-	Problem with initialization.
 *
 *	DESCRIPTION
 *
 *		init_8254() initializes the 8254 timer.  It sets it to the
 *		correct mode, load the default timer values.  and cycles
 *      CNT0H to force the DCC0 output to an initial low state.
 */
/*--------------------------------------------------------------------------*/

int init_8254( void )
{
	/*
	 *	Set 8254 CNT0H to mode 0 to force DCC output low.
	 */
	outp( CNT_CNTRL, CNTRL_0H_RST );
	outp( CNT_0H, CNT_LOWER( CNT_0H_INIT ) );
	outp( CNT_0H, CNT_UPPER( CNT_0H_INIT ) );

	/*
	 *	Set 8254 CNT0T to normal mode and initialize count.
	 */
	outp( CNT_CNTRL, CNTRL_0T_NORM );
	outp( CNT_0T, CNT_LOWER( CNT_0T_INIT ) );
	outp( CNT_0T, CNT_UPPER( CNT_0T_INIT ) );

	/*
	 *	Set 8254 CNT0H to normal mode and initialize count.
	 */
	outp( CNT_CNTRL, CNTRL_0H_NORM );
	outp( CNT_0H, CNT_LOWER( CNT_0H_INIT ) );
	outp( CNT_0H, CNT_UPPER( CNT_0H_INIT ) );

	/*
	 *	Set 8254 CNT1T to normal mode and initialize count.
	 */
	outp( CNT_CNTRL, CNTRL_1T_NORM );
	outp( CNT_1T, CNT_LOWER( CNT_1T_INIT ) );
	outp( CNT_1T, CNT_UPPER( CNT_1T_INIT ) );

	clk0t	=	CNT_0T_INIT;
	clk0h	=	CNT_0H_INIT + 2;
	clk1t	=	CNT_1T_INIT;

	return ( OK );
}


/*--------------------------------------------------------------------------*/
/*
 *	NAME
 *
 *		init_8255()							-	 Init 8255 I/O ports.
 *
 *	RETURN VALUE
 *
 *		OK		-	Sucess.
 *		FAIL	-	Problem with initialization.
 *
 *	DESCRIPTION
 *
 *		init_8255() initializes the 8255 I/O ports.  It sets port A
 *		as a strobed output, port B as a direct output, and port C
 *		as a direct input.  It also initializes port A and B to the
 *		correct initial values.
 */
/*--------------------------------------------------------------------------*/

int init_8255( void )
{
	/*
	 *	Set 8255 to power up condition, all ports as inputs
	 */
	rst_8255();

	/*
	 *	Set 8255 to normal mode.
	 */
	outp( P_CNTRL, P_CNTRL_NORM );

	OUT_PC( PC_POS_BDRST, 1 );				// Reset board.
	OUT_PC( PC_POS_CPUEN, 1 );				// Enable software clock.
	OUT_PC( PC_POS_CPUCLKL, 0 );			// Set SW -CLK to 0.
	OUT_PC( PC_POS_INTRA, 1 );				// Set interrupt active.
	OUT_PC( PC_POS_UNDERCLRL, 1 );			// Set under clear inactive.
	OUT_PC( PC_POS_SCOPENL, 1 );			// Turn off scope signal.
	OUT_PC( PC_POS_ACKAL, 1 );				// Enable interrupt line.
	OUT_PC( PC_POS_OBFAL, 1 );				// Set -OBF to 1.

	OUT_PC( PC_POS_BDRST, 0 );				// Turn off board reset.

	return ( OK );
}


/*--------------------------------------------------------------------------*/
/*
 *	NAME
 *
 *		rst_8255()							-	 Reset 8255 to power up state.
 *
 *	RETURN VALUE
 *
 *		OK		-	Sucess.
 *		FAIL	-	Problem with reset.
 *
 *	DESCRIPTION
 *
 *		rst_8255() resets the 8255 I/O ports to the power on state.
 *		All ports are set to inputs.
 */
/*--------------------------------------------------------------------------*/

int rst_8255( void )
{
	/*
	 *	Set 8255 to power up condition, all ports as inputs
	 */
	outp( P_CNTRL, P_CNTRL_RST );

	return ( OK );
}

/*--------------------------------------------------------------------------*/
/*
 *	NAME
 *
 *		init_send()							-	 Init sender board.
 *
 *	RETURN VALUE
 *
 *		OK		-	Sucess.
 *		FAIL	-	Problem with initialization.
 *
 *	DESCRIPTION
 *
 *		init_send() initializes the DCC sender board.  It sets the timers
 *		and I/O ports for normal operation.  It halts the hardware clock by
 *		placing the board in software clock mode.
 */
/*--------------------------------------------------------------------------*/

int init_send( void )
{
	init_8255();							// Init 8255.
	init_8254();							// Init 8254.
	OUT_PC( PC_POS_CPUCLKL, 1 );			// Toggle clock to load 0T.
	OUT_PC( PC_POS_CPUCLKL, 0 );
	clr_err_cnt();

	return ( OK );
}


/*--------------------------------------------------------------------------*/
/*
 *	NAME
 *
 *		set_clk()							-	 Set DCC clocks.
 *
 *	RETURN VALUE
 *
 *		OK		-	Sucess.
 *		FAIL	-	Problem setting clock.
 *
 *	DESCRIPTION
 *
 *		set_clk() sets one or more of the DCC clocks to a new value.
 *		'iclk0t' sets the 0T time in microseconds.  'iclk0h' sets the
 *		0H time in microseconds.  'iclk1t' sets the 1T time in microseconds.
 *		Setting any of these value to DCC_CLK_HOLD will leave that value
 *		unchanged.
 */
/*--------------------------------------------------------------------------*/

int set_clk(
	u_short			iclk0t,					// 0T duration in usec.
	u_short			iclk0h,					// 0H duration in usec.
	u_short			iclk1t )				// 1T duration in usec.
{
	const char		*my_name = "set_clk";
	unsigned       trunning;
   
   trunning = running;		// Temp running state.

	clr_err_cnt();							// Restart error counter.

	if ( running )							// Stop the clock if it is running.
	   {
		stop_clk();
	   }

	/*
	 *	Set 8254 CNT0T.
	 */
	if ( iclk0t != DCC_CLK_HOLD )
	   {
		if ( iclk0t < DCC_CLK_MIN )
			{
			fprintf(stderr,"%s: iclk0t value %u invalid.\n", my_name, iclk0t );
			}
		else
			{
			clk0t	=	iclk0t;
			outp( CNT_0T, CNT_LOWER( clk0t ) );
			outp( CNT_0T, CNT_UPPER( clk0t ) );
			}
		}

	/*
	 *	Set 8254 CNT0H.  Note - Correct for 2 count delay starting 0H.
	 */
	if ( iclk0h != DCC_CLK_HOLD )
		{
		if ( iclk0h < DCC_CLK_MIN )
			{
			fprintf(stderr,"%s: iclk0h value %u invalid.\n", my_name,iclk0h );
			}
		else
			{
			clk0h	=	iclk0h;
			outp( CNT_0H, CNT_LOWER( clk0h - 2 ) );
			outp( CNT_0H, CNT_UPPER( clk0h - 2 ) );
			}
		}

	/*
	 *	Set 8254 CNT1T.
	 */
	if ( iclk1t != DCC_CLK_HOLD )
		{
		if ( iclk1t < DCC_CLK_MIN )
			{
			fprintf(stderr,"%s: iclk1t value %u invalid.\n", my_name, iclk1t );
			}
		else
			{
			clk1t	=	iclk1t;
			outp( CNT_1T, CNT_LOWER( clk1t ) );
			outp( CNT_1T, CNT_UPPER( clk1t ) );
			}
		}

	/*
	 *	Start clock to flush change through.
	 */
	clr_under();
	start_clk();
	if ( send_bytes( 2, 0x55 ) != OK )
		{
		fprintf(stderr,"%s: send_bytes() FAILED", my_name);
		stop_clk();
		clk0t = clk0h = clk1t = DCC_CLK_ERR;
		return ( FAIL );
		}

	/*
	 *	Reset to the initial run state.
	 */
	if ( !trunning )
		{
		stop_clk();
		}
	
	return ( OK );
}


/*--------------------------------------------------------------------------*/
/*
 *	NAME
 *
 *		start_clk()							-	 Start hardware clock.
 *
 *	RETURN VALUE
 *
 *		OK		-	Sucess.
 *		FAIL	-	Problem starting clock.
 *
 *	DESCRIPTION
 *
 *		start_clk() starts the hardware sender clock.  It does this
 *		in a glitch free way which guarantees that the first hardware
 *		clock pulse will be full width.
 */
/*--------------------------------------------------------------------------*/

int start_clk( void )
{
	const char		*my_name = "start_clk";

	clr_err_cnt();							// Restart error counter.

	if ( running )
		{
		fprintf(stderr,"%s: Clock already running.\n", my_name);
		return ( FAIL );
		}

	OUT_PC( PC_POS_CPUCLKL, 0 );			// Set SW clock to 0.
	OUT_PC( PC_POS_CPUEN, TRUE );			// Stop the HW clock.
	OUT_PC( PC_POS_CPUEN, TRUE );			// Delay so HW clock stops.
	OUT_PC( PC_POS_SCOPENL, 1 );			// Disable scope trigger.
	scope	=	SCOPE_IDLE;
	OUT_PC( PC_POS_CPUEN, FALSE );			// Start glitch free HW clock.
	running	=	TRUE;						// We're off and running.

	return ( OK );
}

/*--------------------------------------------------------------------------*/
/*
 *	NAME
 *
 *		stop_clk()							-	 Stop hardware clock.
 *
 *	RETURN VALUE
 *
 *		OK		-	Sucess.
 *		FAIL	-	Problem stopping clock.
 *
 *	DESCRIPTION
 *
 *		stop_clk() stops the hardware sender clock.  It does this
 *		in a glitch free way which guarantees that the last hardware
 *		clock pulse will be full width.
 */
/*--------------------------------------------------------------------------*/

int stop_clk( void )
{
	const char		*my_name = "stop_clk";

	clr_err_cnt();							// Restart error counter.

	if ( !running )
		{
		fprintf(stderr,"%s: Clock already stopped", my_name);
		return ( FAIL );
		}

	OUT_PC( PC_POS_CPUCLKL, 0 );			// Set SW clock to 0.
	OUT_PC( PC_POS_CPUEN, TRUE );			// Stop the HW clock w/o glitch.
	OUT_PC( PC_POS_CPUEN, TRUE );			// Delay so HW clock stops.
	OUT_PC( PC_POS_SCOPENL, 1 );			// Disable scope trigger.
	scope	=	SCOPE_IDLE;
	running	=	FALSE;						// We're stopped.

	return ( OK );
}

/*--------------------------------------------------------------------------*/
/*
 *	NAME
 *
 *		send_bytes()						-	 Send byte repeatedly.
 *
 *	RETURN VALUE
 *
 *		OK		-	Sucess.
 *		FAIL	-	Problem sending bytes.
 *
 *	DESCRIPTION
 *
 *		send_bytes() sends 'ibyte' 'icnt' times.  It returns FAIL if the
 *		INTRA line does not go inactive in a reasonable length of time or
 *		if the underflow flag gets set.
 */
/*--------------------------------------------------------------------------*/

int send_bytes(
	u_int			icnt,					// Bytes to send.
	BYTE			ibyte )					// Byte to repeat (optional).
{
	const char		*my_name = "send_bytes";
	register u_int	count;					// Count of BYTES sent.
	register u_int	san_cnt;		   		// Sanity timeout.

	if ( icnt < 1 )
		{
		fprintf(stderr,"%s: 0 icnt, b_cnt %lu, p_cnt %lu", my_name, b_cnt, p_cnt );
		return ( FAIL );
		}

	start_crit();
	for ( count = 0; count < icnt; ++count )
		{
		for ( san_cnt =  SAN_CNT;
			san_cnt > 0 && !(inp( PC ) & PC_MSK_INTRA);
			--san_cnt )
			{
			;
			}

		if ( san_cnt ==  0 )
			{
			_enable();
			if ( errprint_ok() )
				{
				fprintf(stderr,"%s: Sanity timeout, sent %u, b_cnt %lu, p_cnt %lu", my_name,
					count, b_cnt, p_cnt );
				}
			return ( FAIL );
			}

		switch ( scope )
			{
		   case SCOPE_IDLE:						// Do nothing.
			   break;

		   case SCOPE_END:							// End scope trigger.
			   OUT_PC( PC_POS_SCOPENL, 1 );
			   scope	=	SCOPE_IDLE;
			   break;

		   case SCOPE_START:						// Start scope trigger.
			   OUT_PC( PC_POS_SCOPENL, 0 );
			   scope	=	SCOPE_END;
			   break;
			}

		if ( inp( PB ) & PB_MSK_UNDER )
			{
			_enable();

			if ( errprint_ok() )
				{
				fprintf(stderr,"%s: Underflow, sent %u, b_cnt %lu, p_cnt %lu", my_name,
					count, b_cnt, p_cnt );
				}
			return ( FAIL );
			}

		outp( PA, ibyte );
		OUT_PC( PC_POS_UNDERCLRL, 1 );
		}

	_enable();
	clr_err_cnt();							// Restart error counter.
	b_cnt	+= count;
	++p_cnt;

	return ( OK );
}


/*--------------------------------------------------------------------------*/
/*
 *	NAME
 *
 *		send_stretched_byte()				-	 Send stretched Byte.
 *
 *	RETURN VALUE
 *
 *		OK		-	Sucess.
 *		FAIL	-	Problem sending byte.
 *
 *	DESCRIPTION
 *
 *		send_stretched byte() sends 'ibyte' with a stretched first 0.
 *		This routine is designed to stretch the interbyte 0 immediately
 *		following the preamble bits by loading clock values 'iclk0t' and
 *		'iclk0h' for the duration of the first 0 bit in 'ibyte'.
 *		It returns FAIL if a problem occurs.
 */
/*--------------------------------------------------------------------------*/

int send_stretched_byte(
	u_short			iclk0t,					// Stretched 0T duration in usec.
	u_short			iclk0h,					// Stretched 0H duration in usec.
	BYTE			ibyte )					// Byte to repeat (optional).
{
	const char		*my_name = "send_stretched_byte";
	register u_int	san_cnt;		   		// Sanity timeout.
	u_short			tclk0t;					// clk0t for stretched 0.

	if ( inp( PA ) != 0xff )
		{
		fprintf(stderr,"%s: PA must be sending preamble bits (all 1s), it is set to <%02x>", my_name,
			inp( PA ) );
		return ( FAIL );
		}

	if ( ibyte & 0x80 )
		{
		fprintf(stderr,"%s: ibyte must have bit 7 set to 0, its value is 0x%02x", my_name,ibyte );
		return ( FAIL );
		}

	if ( iclk0t ==  DCC_CLK_HOLD )
		{
		iclk0t	=	clk0t;
		}

	if ( iclk0h == DCC_CLK_HOLD )
		{
		iclk0h	=	clk0h;
		}

	if ( iclk0t < clk0t )
		{
		fprintf(stderr,"%s: iclk0t value %u invalid, must be >= %u\n", my_name,iclk0t, clk0t );
		return ( FAIL );
		}

	if ( iclk0h < clk0h )
		{
		fprintf(stderr,"%s: iclk0h value %u invalid, must be >= %u\n", my_name, iclk0h, clk0h );
		return ( FAIL );
		}

	/*
	 *	Calculate the CLK0T value to use for the stretched 0.  The hardware
	 *	will use this value for the first half of the total period and the
	 *	original value for the second half.
	 */
	tclk0t	=	(2 * iclk0t) - clk0t;

	start_crit();

	// Step 1: Wait to load 'ibyte' until board is ready.
	for (	san_cnt =  SAN_CNT;
			san_cnt > 0 && !(inp( PC ) & PC_MSK_INTRA);
			--san_cnt )
		{
		;
		}

	if ( san_cnt ==  0 )
		{
		_enable();
		if ( errprint_ok() )
			{
			fprintf(stderr,"%s: Sanity timeout before loading ibyte, b_cnt %lu, p_cnt %lu", my_name,
				b_cnt, p_cnt );
			}
		return ( FAIL );
		}

	switch ( scope )
		{
	   case SCOPE_IDLE:						// Do nothing.
		   break;

	   case SCOPE_END:							// End scope trigger.
		   OUT_PC( PC_POS_SCOPENL, 1 );
		   scope	=	SCOPE_IDLE;
		   break;

	   case SCOPE_START:						// Start scope trigger.
		   OUT_PC( PC_POS_SCOPENL, 0 );
		   scope	=	SCOPE_END;
		   break;
		}

	if ( inp( PB ) & PB_MSK_UNDER )
		{
		_enable();
		if ( errprint_ok() )
			{
			fprintf(stderr,"%s: Underflow, b_cnt %lu, p_cnt %lu\n", my_name,
				b_cnt, p_cnt );
			}
		return ( FAIL );
		}

	outp( PA, ibyte );
	OUT_PC( PC_POS_UNDERCLRL, 1 );

	// Step 2: Disable interrupts and set temporary new 0T and 0H times.
	_disable();

	/*
	 *	Set stretched 8254 CNT0T.
	 */
	outp( CNT_0T, CNT_LOWER( tclk0t ) );
	outp( CNT_0T, CNT_UPPER( tclk0t ) );

	/*
	 *	Set stretched 8254 CNT0H.  Note - Correct for 2 count 0H delay.
	 */
	outp( CNT_0H, CNT_LOWER( iclk0h - 2 ) );
	outp( CNT_0H, CNT_UPPER( iclk0h - 2 ) );

	// Step 3: Wait for board to load 'ibyte'by monitoring INTRA.
	for (	san_cnt =  SAN_CNT;
			san_cnt > 0 && !(inp( PC ) & PC_MSK_INTRA);
			--san_cnt )
		{
		;
		}

	if ( san_cnt ==  0 )
		{
		_enable();
		if ( errprint_ok() )
			{
			fprintf(stderr,"%s: Sanity timeout waiting to load ibyte, b_cnt %lu, p_cnt %lu", my_name,
				b_cnt, p_cnt );
			}
		return ( FAIL );
		}

	// Step 4: Wait for board to start sending 0 bit by monitoring DCCOUTD.
	for (	san_cnt =  SAN_CNT;
			san_cnt > 0 && !(inp( PB ) & PB_MSK_DCCOUTD);
			--san_cnt )
		{
		;
		}

	if ( san_cnt ==  0 )
		{
		_enable();
		if ( errprint_ok() )
			{
			fprintf(stderr,"%s: Sanity timeout before sending first 0, b_cnt %lu, p_cnt %lu", my_name,
				b_cnt, p_cnt );
			}
		return ( FAIL );
		}

	// Step 5: Kill at least 1 usec. while CNT0H gets loaded.
	pc_delay_high( m_pc_delay_1usec );

	// Step 6: Reload original CNT0T and CNT0H values.
	/*
	 *	Reset original 8254 CNT0T.
	 */
	outp( CNT_0T, CNT_LOWER( clk0t ) );
	outp( CNT_0T, CNT_UPPER( clk0t ) );

	/*
	 *	Set stretched 8254 CNT0H.  Note - Correct for 2 count 0H delay.
	 */
	outp( CNT_0H, CNT_LOWER( clk0h - 2 ) );
	outp( CNT_0H, CNT_UPPER( clk0h - 2 ) );

	_enable();
	clr_err_cnt();							// Restart error counter.
	++b_cnt;

	return ( OK );
}

/*--------------------------------------------------------------------------*/
/*
 *	NAME
 *
 *		send_pkt()							-	 Send array of BYTES.
 *
 *	RETURN VALUE
 *
 *		OK		-	Sucess.
 *		FAIL	-	Problem sending bytes.
 *
 *	DESCRIPTION
 *
 *		send_pkt() sends each BYTE in the 'ibytes' array.
 *		It returns FAIL if the INTRA line does not go inactive in a
 *		reasonable length of time or if the underflow flag gets set.
 */
/*--------------------------------------------------------------------------*/

int send_pkt(const BYTE	*ibytes,long lsize )
{
	const char		*my_name = "send_pkt(BYTE)";
	register unsigned long	lcount; 		// Count of BYTES sent.
	register u_int	san_cnt;		   		// Sanity timeout.
   BYTE *pBytes;

	if ( lsize < 1L )
		{
		fprintf(stderr,"%s: 0 lsize, b_cnt %lu, p_cnt %lu", my_name, b_cnt, p_cnt );
		return ( FAIL );
		}


// Version 0.95
// Commented Out Print Statment
//    printf("\nlsize = %lu\n",lsize);


	/* Send first BYTE, then activate underflow warning */

	start_crit();
	for ( lcount = 0L; lcount < lsize; ++ibytes, ++lcount )
		{
		for ( san_cnt =  SAN_CNT;
			san_cnt > 0 && !(inp( PC ) & PC_MSK_INTRA);
			--san_cnt )
			{
			;
			}

		if ( san_cnt ==  0 )
 			{
			_enable();
			if ( errprint_ok() )
				{
				fprintf(stderr,"%s: Sanity timeout, sent %lu, b_cnt %lu, p_cnt %lu\n", my_name,
					lcount, b_cnt, p_cnt );
				}
			return ( FAIL );
			}

		switch ( scope )
			{
		   case SCOPE_IDLE:						// Do nothing.
			   break;

		   case SCOPE_END:							// End scope trigger.
			   OUT_PC( PC_POS_SCOPENL, 1 );
			   scope	=	SCOPE_IDLE;
			   break;

		   case SCOPE_START:						// Start scope trigger.
			   OUT_PC( PC_POS_SCOPENL, 0 );
			   scope	=	SCOPE_END;
			   break;
			}

		if ( inp( PB ) & PB_MSK_UNDER )
			{
			_enable();

         printf("\nPB= %X, PB_MSK_UNDER = %X\n", inp(PB),PB_MSK_UNDER);

			if ( errprint_ok()  )
				{
				fprintf(stdout,"%s: Underflow, sent %lu, b_cnt %lu, p_cnt %lu\n", my_name,
					 lcount, b_cnt, p_cnt );
				}
			return ( FAIL );
			}

		outp( PA, *ibytes );
		OUT_PC( PC_POS_UNDERCLRL, 1 );
		}

	_enable();
	clr_err_cnt();							// Restart error counter.
	b_cnt	+= lcount;
	++p_cnt;

	return ( OK );
}

/*--------------------------------------------------------------------------*/
/*
 *	NAME
 *
 *		errprint_ok()					   	-	 Check if OK to print error.
 *
 *	RETURN VALUE
 *
 *		TRUE	-	OK to print error.
 *		FALSE	-	Don't print the error.
 *
 *	DESCRIPTION
 *
 *		errprint_ok() return TRUE if the number of contiguous errors
 *		given by 'err_cnt' has not been exceeded.  It also prints an
 *		error of its own when 'err_cnt' is just reached.
 */
/*--------------------------------------------------------------------------*/

int errprint_ok( void )
{
	const char		*my_name = "errprint_ok";

	if ( err_cnt == ERR_MAX )
		{
		fprintf(stderr,"%s: Contiguous error count of %u reached\n", my_name, err_cnt );
		++err_cnt;
		}
	else if ( err_cnt < ERR_MAX  )
		{
		++err_cnt;
		return ( TRUE );
		}

	return ( FALSE );
}

/*--------------------------------------------------------------------------*/
/*
 *	NAME
 *
 *		print_stats()						-	 Print packet statistics.
 *
 *	RETURN VALUE
 *
 *		NONE
 *
 *	DESCRIPTION
 *
 *		print_stats() prints packet sending statistics to stdout.
 */
/*--------------------------------------------------------------------------*/

void print_stats( void )
{
	printf(
		"Packets sent %lu, Bytes sent %lu\n", p_cnt, b_cnt );
}

/*--------------------------------------------------------------------------*/
/*
 *	NAME
 *
 *		set_pc_delay_1usec()	   	 	-	 Set PC 1 usec delay value.
 *
 *	RETURN VALUE
 *
 *		OK		-	Sucess.
 *		FAIL	-	Problem setting m_pc_delay_1usec.
 *
 *	DESCRIPTION
 *
 *		set_pc_delay_1usec tests the PC 1 usec delay time and sets
 *		'm_pc_delay_1usec' accordingly.  It returns FAIL if a problem
 *		occurs.
 */
/*--------------------------------------------------------------------------*/

int set_pc_delay_1usec( void )
{
	const char		*my_name = "set_pc_delay_1usec";
	register u_int	san_cnt;		   		// Sanity timeout.
	register u_int	tmp_delay_high;	 		// pc_delay_high() return value.
	register u_int	tmp_delay_low;	 		// pc_delay_low() return value.

	/*
	 *	Initialize the hardware and start the clock at nominal values.
	 */
	init_send();
	if ( set_clk(	DECODER_0T_NOM,
					DECODER_0H_NOM,
					DECODER_1T_NOM ) != OK )
		{
		fprintf(stderr,"%s: set_clk() FAILED\n", my_name );
		return ( FAIL );
		}
	start_clk();

	if ( send_bytes( 1, 0x00 ) != OK )
		{
		fprintf(stderr,"%s: send_bytes() FAILED\n", my_name );
		stop_clk();
		return ( FAIL );
		}

	// Step 1: Disable interrupts.
	_disable();

	// Step 2: Wait for board to load 'ibyte'by monitoring INTRA.
	for (	san_cnt =  SAN_CNT;
			san_cnt > 0 && !(inp( PC ) & PC_MSK_INTRA);
			--san_cnt )
		{
		;
		}

	if ( san_cnt ==  0 )
		{
		_enable();
		if ( errprint_ok() )
			{
			fprintf(stderr,"%s: Sanity timeout waiting to load ibyte", my_name );
			}
		return ( FAIL );
		}

	// Step 3: Wait for board to start sending 0 bit by monitoring DCCOUTD.
	for (	san_cnt =  SAN_CNT;
			san_cnt > 0 && !(inp( PB ) & PB_MSK_DCCOUTD);
			--san_cnt )
		{
		;
		}

	if ( san_cnt ==  0 )
		{
		_enable();
		if ( errprint_ok() )
			{
			fprintf(stderr,"%s: Sanity timeout before sending first 0", my_name);
			}
		return ( FAIL );
		}

	// Step 4: Time the high period of DCCOUTD.
	tmp_delay_high	=	pc_delay_high( SAN_CNT );

	// Step 5: Time the low period of DCCOUTD.
	tmp_delay_low	=	pc_delay_low( SAN_CNT );

	// Step 6: enable interrupts, stop the hardware, and evaluate the results.
	_enable();
	stop_clk();

	printf(		"PC high %4u(%4u), low %4u(%4u)\n",
		clk0h, SAN_CNT - tmp_delay_high,
		clk0t - clk0h, SAN_CNT - tmp_delay_low );

	m_pc_delay_1usec	=	((SAN_CNT - tmp_delay_high)/DECODER_0H_NOM) + 1;
	printf(		"PC 1 usec delay %u\n", m_pc_delay_1usec );

	return ( OK );	}

/*--------------------------------------------------------------------------*/
/*
 *	NAME
 *
 *		pc_delay_high()	   		 		-	 PC driven delay for DCCOUT high.
 *
 *	RETURN VALUE
 *
 *		idelay	-	Time remaining, if any.
 *
 *	DESCRIPTION
 *
 *		pc_delay_high() can test or generate a PC driven delay.
 *		This method assumes that interrupts are disabled
 *		and that DCCOUTD is high prior to the call.  This allows the same
 *		method to be called from set_pc_delay_1usec() to find the correct
 *		'idelay' value and then to be called from other methods to
 *		generate the delay.
 */
/*--------------------------------------------------------------------------*/

u_int pc_delay_high(
	register u_int	idelay )			// Delay count.
{
	while ( idelay > 0 && (inp( PB ) & PB_MSK_DCCOUTD) )
		{
		idelay--;
		}
	return ( idelay );
}


/*--------------------------------------------------------------------------*/
/*
 *	NAME
 *
 *		pc_delay_low()	   		 		-	 PC driven delay for DCCOUT low.
 *
 *	RETURN VALUE
 *
 *		idelay	-	Time remaining, if any.
 *
 *	DESCRIPTION
 *
 *		pc_delay_low() can test or generate a PC driven delay.
 *		This method assumes that interrupts are disabled
 *		and that DCCOUTD is low prior to the call.  This allows the same
 *		method to be called from set_pc_delay_1usec() to find the correct
 *		'idelay' value and then to be called from other methods to
 *		generate the delay.
 */
/*--------------------------------------------------------------------------*/

u_int pc_delay_low(
	register u_int	idelay )			// Delay count.
{
	while ( idelay > 0 && !(inp( PB ) & PB_MSK_DCCOUTD) )
		{
		idelay--;
		}
	return ( idelay );
}


